import json
import logging
import os

logger = logging.getLogger(__name__)


class MessageHeader:
    """
    Descriptor to map a STOMP message header name to a Python attribute.
    """
    def __init__(self, header_name: str):
        self.header_name = header_name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if not hasattr(instance, 'headers'):
            raise TypeError(f'Expected {instance} to have a "headers" attribute')
        return instance.headers.get(self.header_name)

    def __set__(self, instance, value):
        if not hasattr(instance, 'headers'):
            raise TypeError(f'Expected {instance} to have a "headers" attribute')
        instance.headers[self.header_name] = value

    def __delete__(self, instance):
        if not hasattr(instance, 'headers'):
            raise TypeError(f'Expected {instance} to have a "headers" attribute')
        if self.header_name in instance.headers:
            del instance.headers[self.header_name]


class Message:
    id = MessageHeader('message-id')
    persistent = MessageHeader('persistent')

    @classmethod
    def read(cls, filename):
        headers = {}
        body = ''
        in_body = False
        with open(filename, 'r') as fh:
            for line in fh:
                if in_body:
                    body += line
                else:
                    if line.rstrip() == '':
                        in_body = True
                        continue
                    else:
                        (key, value) = line.split(':', 1)
                        headers[key] = value.strip()
        return cls(headers=headers, body=body)

    def __init__(self, message_id=None, persistent=None, headers=None, body=''):
        if headers is not None:
            self.headers = headers
        else:
            self.headers = {}

        if isinstance(body, dict):
            # convert dictionaries to JSON
            self.body = json.dumps(body)
        else:
            # treat all other bodies as strings
            self.body = str(body)

        if message_id is not None:
            self.id = message_id
        if persistent is not None:
            self.persistent = persistent

    def __str__(self):
        return '\n'.join([f'{k}:{v}' for k, v in self.headers.items()]) + '\n\n' + self.body


class PlastronMessage(Message):
    job_id = MessageHeader('PlastronJobId')

    def __init__(self, job_id: str = None, **kwargs):
        super().__init__(**kwargs)
        # Plastron message are persistent by default
        if 'persistent' not in self.headers:
            self.persistent = 'true'
        if job_id is not None:
            self.job_id = job_id


class PlastronResponseMessage(PlastronMessage):
    state = MessageHeader('PlastronJobState')

    def __init__(self, state: str = None, **kwargs):
        super().__init__(**kwargs)
        if state is not None:
            self.state = state


class PlastronErrorMessage(PlastronMessage):
    error = MessageHeader('PlastronJobError')

    def __init__(self, error: str = None, **kwargs):
        super().__init__(**kwargs)
        if error is not None:
            self.error = error


class PlastronCommandMessage(PlastronMessage):
    command = MessageHeader('PlastronCommand')

    def __init__(self, command: str = None, args: dict = None, **kwargs):
        super().__init__(**kwargs)
        if command is not None:
            self.command = command
        if args is not None:
            for name, value in args.items():
                self.headers[f'PlastronArg-{name}'] = value

    @property
    def args(self):
        return {h[12:]: v for h, v in self.headers.items() if h.startswith('PlastronArg-')}

    def response(self, state: str, body) -> PlastronResponseMessage:
        """
        Return a new PlastronResponseMessage with the same job_id as this message,
        plus the given state and body.
        """
        return PlastronResponseMessage(job_id=self.job_id, state=state, body=body)


class MessageBox:
    def __init__(self, directory, message_class=None):
        self.dir = directory
        os.makedirs(directory, exist_ok=True)
        self.message_class = message_class

    def add(self, message_id, message):
        filename = os.path.join(self.dir, message_id.replace('/', '-'))
        with open(filename, 'wb') as fh:
            fh.write(str(message).encode())

    def remove(self, message_id):
        filename = os.path.join(self.dir, message_id.replace('/', '-'))
        os.remove(filename)

    def __iter__(self):
        self.filenames = os.listdir(self.dir)
        self.index = 0
        self.count = len(self.filenames)
        return self

    def __next__(self):
        if self.index < self.count:
            filename = os.path.join(self.dir, self.filenames[self.index])
            self.index += 1
            if self.message_class is not None:
                return self.message_class.read(filename)
            else:
                # just return the filename if no message class was defined
                return filename
        else:
            raise StopIteration



import logging
from pathlib import Path
from typing import NamedTuple, Optional, Dict, Union

from stomp import Connection11
from stomp.exception import StompException

from plastron.messaging.messages import Message

logger = logging.getLogger(__name__)


class ServerTuple(NamedTuple):
    host: str
    port: int

    @classmethod
    def from_string(cls, value: str) -> 'ServerTuple':
        host, port = value.split(':', 1)
        return cls(host=host, port=int(port))

    def __str__(self):
        return f'{self.host}:{self.port}'


class Broker:
    def __init__(
            self,
            server: ServerTuple,
            message_store_dir: Union[Path, str],
            destinations: Optional[Dict[str, str]] = None,
            public_uri_template: Optional[str] = None,
    ):
        self.server = server
        self.connection = Connection11([self.server])
        self.destinations = {key.upper(): Destination(self, value) for key, value in destinations.items()}
        self.message_store_dir = message_store_dir
        self.public_uri_template = public_uri_template
        self.client_id = None

    def __str__(self):
        return f'{self.server} (client-id: {self.client_id})'

    def __getitem__(self, item) -> 'Destination':
        return self.destination(item)

    def connect(self, client_id: str) -> bool:
        if not self.connection.is_connected():
            logger.info(
                f'Attempting to connect to STOMP message broker ('
                f'Host: {self.server[0]}, '
                f'Port: {self.server[1]}, '
                f'Client ID: {client_id})'
            )
            try:
                self.connection.connect(wait=True, headers={'client-id': client_id})
            except StompException:
                logger.error(f'STOMP connection failed for {self}')
                return False
            else:
                self.client_id = client_id
        return self.connection.is_connected()

    def disconnect(self):
        self.connection.disconnect()
        self.client_id = None

    def set_listener(self, *args):
        self.connection.set_listener(*args)

    def subscribe(self, *args, **kwargs):
        self.connection.subscribe(*args, **kwargs)

    def ack(self, *args):
        self.connection.ack(*args)

    def destination(self, name: str) -> 'Destination':
        return self.destinations[name.upper()]

    def send(self, destination, headers=None, body='', **kwargs):
        if headers is None:
            headers = {}
        self.connection.send(
            destination=destination,
            headers=headers,
            body=body,
            **kwargs
        )


class Destination:
    def __init__(self, broker: Broker, destination: str):
        self.broker = broker
        self.name = destination

    def __str__(self):
        return self.name

    def send(self, message: Message):
        self.broker.connection.send(destination=self.name, headers=message.headers, body=message.body)

    def subscribe(self, id: str, ack: str = 'auto', headers: Dict = None, **kwargs):
        self.broker.connection.subscribe(destination=self.name, id=id, ack=ack, headers=headers, **kwargs)
        logger.info(f"Subscribed to {self.name}")
        logger.debug(f"id={id} ack={ack} headers={headers} {kwargs}")



from plastron.messaging.messages import PlastronCommandMessage, PlastronErrorMessage, PlastronMessage, \
    PlastronResponseMessage


def test_plastron_message_no_body():
    msg = PlastronMessage(message_id='foo', job_id='1')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.persistent == 'true'
    assert msg.body == ''
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['persistent'] == 'true'


def test_plastron_message_string_body():
    msg = PlastronMessage(message_id='foo', job_id='1', body='hello world')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.persistent == 'true'
    assert msg.body == 'hello world'
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['persistent'] == 'true'


def test_plastron_message_dict_body():
    msg = PlastronMessage(message_id='foo', job_id='1', body={'greeting': 'hello world'})
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.persistent == 'true'
    assert msg.body == '{"greeting": "hello world"}'
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['persistent'] == 'true'


def test_plastron_message_not_persistent():
    msg = PlastronMessage(message_id='foo', job_id='1', persistent='false')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.persistent == 'false'
    assert msg.body == ''
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['persistent'] == 'false'


def test_plastron_response_message():
    msg = PlastronResponseMessage(state='Done', message_id='foo', job_id='1')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.state == 'Done'
    assert msg.persistent == 'true'
    assert msg.body == ''
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['PlastronJobState'] == 'Done'
    assert msg.headers['persistent'] == 'true'


def test_plastron_error_message():
    msg = PlastronErrorMessage(error='File Not Found', message_id='foo', job_id='1')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.error == 'File Not Found'
    assert msg.persistent == 'true'
    assert msg.body == ''
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['PlastronJobError'] == 'File Not Found'
    assert msg.headers['persistent'] == 'true'


def test_plastron_command_message():
    msg = PlastronCommandMessage(command='run', message_id='foo', job_id='1')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.command == 'run'
    assert msg.persistent == 'true'
    assert msg.body == ''
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['PlastronCommand'] == 'run'
    assert msg.headers['persistent'] == 'true'


def test_plastron_command_message_with_args():
    msg = PlastronCommandMessage(command='run', args={'size': 38, 'color': 'blue'}, message_id='foo', job_id='1')
    assert msg.id == 'foo'
    assert msg.job_id == '1'
    assert msg.command == 'run'
    assert msg.args == {'size': 38, 'color': 'blue'}
    assert msg.persistent == 'true'
    assert msg.body == ''
    assert msg.headers['message-id'] == 'foo'
    assert msg.headers['PlastronJobId'] == '1'
    assert msg.headers['PlastronCommand'] == 'run'
    assert msg.headers['PlastronArg-size'] == 38
    assert msg.headers['PlastronArg-color'] == 'blue'
    assert msg.headers['persistent'] == 'true'
